using System.Collections;
using UnityEngine;
using UnityEngine.SceneManagement;
using TMPro;
using UnityEngine.Events;

public class Player : MonoBehaviour {
    #region Animation
    private Animator _animator;
    private SpriteRenderer _spriteRenderer;
    #endregion

    #region Climbing
    [HideInInspector]
    public bool CanClimb = false;

    [HideInInspector]
    public bool BottomLadder = false;

    [HideInInspector]
    public bool TopLadder = false;
    private const float _climBSpeed = 3F;
    public Ladder Ladder;
    private float _normalGravity;
    #endregion

    #region What is Ground
    [SerializeField]
    private LayerMask _groundLayer;
    private const float _groundedRadius = 0.2f;

    [SerializeField]
    private Transform _checkGround;
    private bool _isGrounded;
    #endregion

    #region Movement
    private const float _moveSpeed = 25f;
    private float _horizontalMove = 0f;
    private Vector3 _velocity = Vector3.zero;
    private Vector3 _scale = Vector2.one;
    private float _jumpForce = 650.0F;
    private Rigidbody2D _rigidBody2D;
    private Collider2D _collider2D;
    private bool flipX = true;
    #endregion

    #region AudioSources
    [SerializeField] private AudioSource _footstepsSound;

    [SerializeField] private AudioSource _jumpSound;

    [SerializeField] private AudioSource _hurtSound;

    [SerializeField] private AudioSource _shootSound;
    #endregion

    #region Collectable Items
    public int Ammoes = 2;
    public int HitPoint = 5;
    public int Coins = 0;

    [SerializeField] private TextMeshProUGUI _ammoesText;
    [SerializeField] private TextMeshProUGUI _coinsText;
    [SerializeField] private TextMeshProUGUI _jumpText;
    private Bullet _bullet;
    #endregion

    #region HP
    private LivesBar _liveBar;
    public int HP {
        get { return HitPoint; }
        set {
            if (value <= 5)
                HitPoint = value;
        }
    }
    #endregion

    [SerializeField] private ParticleSystem _dustParticle;
    [SerializeField] private ParticleSystem _damageParticle;

    [SerializeField] private TextMeshProUGUI _notEnoughCoinsText;
    [SerializeField] private TextMeshProUGUI _introduction;

    private SaveSystem _saveSystem;

    public static Player SingletonInstance { get; private set; }

    [SerializeField] private UnityEvent OnLand;

    private void Awake() {
        SingletonInstance = this;
        OnLand ??= new UnityEvent();
    }

    private void Start() {
        _liveBar = LivesBar.SingletonInstance;
        _saveSystem = SaveSystem.SingletonInstance;
        _rigidBody2D = GetComponent<Rigidbody2D>();
        _animator = GetComponent<Animator>();
        _spriteRenderer = GetComponent<SpriteRenderer>();
        _bullet = Resources.Load<Bullet>(nameof(Bullet));
        _collider2D = GetComponent<Collider2D>();
        _normalGravity = _rigidBody2D.gravityScale;
        HitPoint = 5;
        ResetDamageResistance();
        StartCoroutine(ResetIntroductionText());
    }

    private void Update() {
        _horizontalMove = Input.GetAxisRaw("Horizontal") * _moveSpeed;
        if (Input.GetButtonDown("Jump") && _isGrounded) Jump();
        if (Input.GetButtonDown("Fire1") && Ammoes > 0) Shoot();
        if (HP <= 0) {
            if (_saveSystem.CanSave) SceneManager.LoadScene(0);
            else _saveSystem.Load();
        }
        _ammoesText.text = Ammoes.ToString();
        _coinsText.text = Coins.ToString();
        _liveBar.Refresh();
        Animations();
    }

    private void FixedUpdate() {
        bool notOnGround = _isGrounded;
        _isGrounded = false;
        Collider2D[] colliders = Physics2D.OverlapCircleAll(
            _checkGround.position,
            _groundedRadius,
            _groundLayer
        );
        for (int i = 0; i < colliders.Length; i++) {
            if (colliders[i].gameObject != gameObject) {
                _isGrounded = true;
                if (!notOnGround)
                    OnLand.Invoke();
            }
        }
        Movement();
        TextFollow();
    }

    private void TextFollow() {
        Vector2 textPos = new Vector2(transform.position.x, transform.position.y + 2f);
        _introduction.transform.position = textPos;
    }

    private void LoadControl() {
        if (HitPoint <= 0 && _saveSystem.CanSave)
            SceneManager.LoadScene(0);
        else
            _saveSystem.Load();
    }

    private void Animations() {
        _animator.SetFloat("SpeedY", _rigidBody2D.velocity.y);
        _animator.SetFloat("SpeedX", Mathf.Abs(_horizontalMove));
        if (_animator.GetBool("Climbing") == true)
            Climb();
        if (_rigidBody2D.velocity.y < 0.1f && !_isGrounded && !CanClimb)
            _animator.SetBool("Falling", true);
    }

    private void Movement() {
        LadderFixation();
        Run(_horizontalMove * Time.fixedDeltaTime);
    }

    private void Run(float moveSpeed) {
        Vector3 targetVelocity = new Vector2(moveSpeed * 10f, _rigidBody2D.velocity.y);
        _rigidBody2D.velocity = Vector3.SmoothDamp(_rigidBody2D.velocity, targetVelocity, ref _velocity, 0f);
        if (_horizontalMove > 0f && !flipX) Facing();
        else if (_horizontalMove < 0f && flipX) Facing();
    }

    private void Facing() {
        flipX = !flipX;
        _scale = transform.localScale;
        _scale.x *= -1f;
        transform.localScale = _scale;
        DustPlay();
    }

    private void DustPlay() => _dustParticle.Play();

    public void IsOnLandEvent() {
        _animator.SetBool("Falling", false);
        _animator.SetBool("Jumping", false);
    }

    private void Jump() {
        _rigidBody2D.AddForce(Vector2.up * _jumpForce);
        _animator.SetBool("Jumping", true);
        _jumpSound.Play();
        DustPlay();
    }

    private void Shoot() {
        Ammoes--;
        _shootSound.Play();
        Bullet newBullet = Instantiate(_bullet, transform.position, _bullet.transform.rotation) as Bullet;
        newBullet.Parent = gameObject;
        newBullet.Direction = newBullet.transform.right * _scale.x;
    }

    private void ReceiveDamage() {
        HP--;
        _hurtSound.Play();
        _damageParticle.Play();
        Physics2D.IgnoreLayerCollision(3, 7, true);
        Physics2D.IgnoreLayerCollision(10, 7, true);
        StartCoroutine(DamageFx());
    }

    public void PowerUpEvent() {
        _jumpText.enabled = true;
        _jumpForce = 1000F;
        StartCoroutine(ResetPowerUp());
    }

    private void LadderFixation() {
        if (CanClimb && Mathf.Abs(Input.GetAxis("Vertical")) > 0.1F) {
            _animator.SetBool("Climbing", true);
            _animator.SetBool("Falling", false);
            _animator.SetBool("Jumping", false);
            _rigidBody2D.constraints =
                RigidbodyConstraints2D.FreezePositionX | RigidbodyConstraints2D.FreezeRotation;
            transform.position = new Vector3(Ladder.transform.position.x, _rigidBody2D.position.y);
            _rigidBody2D.gravityScale = 0F;
        }
    }

    private void Climb() {
        if (Input.GetButtonDown("Jump")) {
            _rigidBody2D.constraints = RigidbodyConstraints2D.FreezeRotation;
            CanClimb = false;
            _animator.SetBool("Climbing", false);
            _rigidBody2D.gravityScale = _normalGravity;
            _animator.speed = 1f;
            Jump();
            return;
        }
        float vDirection = Input.GetAxis("Vertical");
        if (vDirection > 0.1F && !TopLadder) {
            _rigidBody2D.velocity = new Vector2(0F, vDirection * _climBSpeed);
            _animator.speed = 1f;
        } else if (vDirection < -0.1F && !BottomLadder) {
            _rigidBody2D.velocity = new Vector2(0F, vDirection * _climBSpeed);
            _animator.speed = 1f;
        } else {
            _rigidBody2D.velocity = Vector2.zero;
            _animator.speed = 0f;
        }
    }

    private void Footstep() => _footstepsSound.Play();

    private void OnTriggerEnter2D(Collider2D collider) {
        if (collider.TryGetComponent(out Bullet bullet) && bullet.Parent != gameObject)
            ReceiveDamage();

        if (collider.CompareTag("Finish")) {
            if (Coins >= 25)
                SceneManager.LoadScene(1);
            else
                StartCoroutine(ResetText());
        }

        if (collider.CompareTag("HitPoint") && HP == 5)
            return;
        else if (collider.TryGetComponent(out CollectableItems collectable))
            collectable.PickUpEvent();

        if (collider.CompareTag("DeathZone")) {
            if (_saveSystem.CanSave)
                SceneManager.LoadScene(0);
            else
                _saveSystem.Load();
        }
    }

    private void OnCollisionEnter2D(Collision2D other) {
        if (other.gameObject.TryGetComponent(out EnemiesController unit)) {
            ContactPoint2D point = other.contacts[1];
            if (point.normal.y >= 1f) {
                unit.JumpedOn();
                Jump();
            } else
                ReceiveDamage();
        }
        if (other.gameObject.layer.Equals(10))
            ReceiveDamage();
        if (other.gameObject.layer.Equals(9))
            this.transform.parent = other.transform;
    }

    private void OnCollisionStay2D(Collision2D other) {
        if (other.gameObject.TryGetComponent(out EnemiesController unit))
            ReceiveDamage();
        if (other.gameObject.layer.Equals(10))
            ReceiveDamage();
    }

    private void OnCollisionExit2D(Collision2D other) {
        if (other.gameObject.layer.Equals(9)) this.transform.parent = null;
    }

    private IEnumerator DamageFx() {
        _spriteRenderer.color = Color.red;
        yield return new WaitForSeconds(0.25F);
        _spriteRenderer.color = Color.white;
        yield return new WaitForSeconds(0.25F);
        _spriteRenderer.color = Color.red;
        yield return new WaitForSeconds(0.25F);
        _spriteRenderer.color = Color.white;
        yield return new WaitForSeconds(0.25F);
        ResetDamageResistance();
    }

    private void ResetDamageResistance() {
        Physics2D.IgnoreLayerCollision(3, 7, false);
        Physics2D.IgnoreLayerCollision(10, 7, false);
    }

    private IEnumerator ResetText() {
        _notEnoughCoinsText.enabled = true;
        yield return new WaitForSeconds(5f);
        _notEnoughCoinsText.enabled = false;
    }

    private IEnumerator ResetIntroductionText() {
        yield return new WaitForSeconds(8f);
        _introduction.enabled = false;
    }

    private IEnumerator ResetPowerUp() {
        _spriteRenderer.color = Color.magenta;
        yield return new WaitForSeconds(1.5f);
        _spriteRenderer.color = Color.white;
        yield return new WaitForSeconds(0.15f);
        _spriteRenderer.color = Color.magenta;
        yield return new WaitForSeconds(0.15f);
        _spriteRenderer.color = Color.white;
        yield return new WaitForSeconds(0.15f);
        _spriteRenderer.color = Color.magenta;
        yield return new WaitForSeconds(0.15f);
        _spriteRenderer.color = Color.white;
        yield return new WaitForSeconds(0.15f);
        _spriteRenderer.color = Color.magenta;
        yield return new WaitForSeconds(0.15f);
        _spriteRenderer.color = Color.white;
        _jumpForce = 650F;
        _jumpText.enabled = false;
    }
}